/*

  Magia DNI - Calculate the check digit of Spanish identity cards.
  Copyright (C) 2011-2024 Josep Portella Florit <hello@josep-portella.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

const NUMBERS_ROW = 0
const NUMBERS_COLUMN = 5

const NIF_NUMBER_LEN = 8
const NIF_LEN = NIF_NUMBER_LEN + 1

const DATE_LEN = 6

const DNI_NIF_COLUMN = NUMBERS_COLUMN
const DNI_NIF_LETTER_COLUMN = DNI_NIF_COLUMN + NIF_NUMBER_LEN
const DNI_NIF_END_COLUMN = DNI_NIF_LETTER_COLUMN + 1 + 1

const DNIE_SERIAL_NUMBER_COLUMN = NUMBERS_COLUMN
const DNIE_SERIAL_NUMBER_LEN = 9
const DNIE_SERIAL_NUMBER_LETTERS_END_COLUMN = NUMBERS_COLUMN + 3
const DNIE_SERIAL_NUMBER_CHECK_DIGIT_COLUMN
      = DNIE_SERIAL_NUMBER_COLUMN + DNIE_SERIAL_NUMBER_LEN
const DNIE_NIF_COLUMN = DNIE_SERIAL_NUMBER_CHECK_DIGIT_COLUMN + 1
const DNIE_NIF_LETTER_COLUMN = DNIE_NIF_COLUMN + NIF_NUMBER_LEN
const DNIE_LAST_DIGIT_COLUMN = DNIE_NIF_LETTER_COLUMN - 1
const DNIE_NIF_END_COLUMN = DNIE_NIF_COLUMN + NIF_LEN

const DATES_ROW = 1
const BIRTHDATE_COLUMN = 0
const BIRTHDATE_CHECK_DIGIT_COLUMN = BIRTHDATE_COLUMN + DATE_LEN
const EXPIRATION_DATE_COLUMN = BIRTHDATE_CHECK_DIGIT_COLUMN + 1 + 1
const EXPIRATION_DATE_CHECK_DIGIT_COLUMN = EXPIRATION_DATE_COLUMN + DATE_LEN

const NUMBER_OF_COLUMNS = DNIE_NIF_END_COLUMN

const DIGITS = '0123456789'
const DIGITS_OR_NULL = DIGITS + '<'
const NIF_LETTERS = 'TRWAGMYFPDXBNJZSQVHLCKE'
const LETTERS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

const CHAR_CODE_0 = '0'.charCodeAt()
const CHAR_CODE_A = 'A'.charCodeAt()

async function onLoad(update) {
    const templatesPromise = loadTemplates()

    displayMessage('Inicializando la cámara...')
    hideOverlay()

    document.body.addEventListener('click', event => {
        if (event.target.tagName != 'A'
            && event.target.parentElement.tagName != 'A') {
            if (!document.fullscreenElement)
                document.body.requestFullscreen().then(() =>
                    screen.orientation.lock('landscape'))
            else if (document.exitFullscreen)
                document.exitFullscreen()
        }
    })

    if (navigator.wakeLock)
        try {
            await navigator.wakeLock.request('screen')
        } catch (error) {}

    var stream
    if (navigator.mediaDevices)
        try {
            stream = await navigator.mediaDevices.getUserMedia({
                video: {width: 640, facingMode: 'environment'}})
        } catch (error) {}

    if (!stream)
        return displayMessage('No se pudo inicializar la cámara.')

    displayMessage('Cargando...')
    try {
        var templates = await templatesPromise
    } catch (error) {
        return displayMessage('No se pudieron cargar las plantillas.')
    }

    main(templates, stream, update)
}

async function main(templates, stream, update) {
    const video = document.querySelector('video')
    video.srcObject = stream

    var checkDigit

    update = update || function (ocrData) {
        if (ocrData.rows && !ocrData.columns)
            displayRows(ocrData)

        if (ocrData.checkDigit && checkDigit === ocrData.checkDigit) {
            displayCheckDigit(checkDigit)
            checkDigit = ''
            displayMessage('')
            hideOverlay()
            return 4000
        }

        if (ocrData.columns)
            displaySignificantChars(ocrData)

        if (checkDigit = ocrData.checkDigit) {
            displayMessage('No muevas la cámara...')
            return 500
        }
    }

    function loop() {
        displayMessage('Encaja en el rectángulo las 3 líneas de caracteres '
                       + 'OCR del DNI,<br> iluminadas adecuadamente. '
                       + 'Aleja y acerca la cámara para enfocar.')
        displayCheckDigit('')
        showOverlay()
        clear()
        const imageData = getVideoImageData(video)
        setTimeout(loop, imageData
                   ? update(new OcrData(
                       templates, imageData, .2, innerWidth, innerHeight))
                   : displayMessage(
                       'Pon el dispositivo en orientación horizontal.'))
    }

    loop()
}

async function loadTemplates() {
    const templates = {}
    for (const c of DIGITS_OR_NULL + LETTERS)
        templates[c] = loadImageData(`template/${c}.gif`)
    for (const c in templates)
        templates[c] = await templates[c]
    return templates
}

const DNI_FORMAT = {
    FIELD_GETTERS: ['getOldNif'],

    getPossibleChars(column) {
        return column == DNI_NIF_LETTER_COLUMN ? NIF_LETTERS : DIGITS
    },

    isSignificantChar(column) {
        return column < DNI_NIF_END_COLUMN
    },
}

const DNIE_FORMAT = {
    FIELD_GETTERS: ['getSerialNumber', 'getNif'],

    getPossibleChars(column) {
        if (column >= DNIE_SERIAL_NUMBER_COLUMN
            && column < DNIE_SERIAL_NUMBER_LETTERS_END_COLUMN)
            return LETTERS
        if (column == DNIE_NIF_LETTER_COLUMN)
            return NIF_LETTERS
        return DIGITS
    },

    isSignificantChar(column) {
        return column < DNIE_NIF_END_COLUMN
    },
}

class OcrData {
    constructor(templates, imageData, widthFactor, viewWidth, viewHeight) {
        const timestamp = Date.now()

        this.width = imageData.width
        this.height = Math.round(this.width * widthFactor)

        if (!viewWidth)  viewWidth  = imageData.width
        if (!viewHeight) viewHeight = imageData.height

        this.ratio = imageData.width / viewWidth
        this.fullWidth = imageData.width

        this.xOffset = 0
        this.yOffset = Math.round(
            this.ratio * (viewHeight - viewWidth * widthFactor) / 2)

        this.rgba = imageData.data
        this.templates = templates

        this.minRowSize = this.height / 10
        this.rowValues = Array.from(this.calcRowValues())

        if (this.rows = this.findRows()) {
            this.columnValues = Array.from(this.calcColumnValues())
            this.minColumnSize = ((this.rows[1].position + this.rows[1].size
                                   - this.rows[0].position)
                                  / 8)
            if (this.columns = this.findColumns()) {
                this.format = this.getFormat()
                this.fields = this.getFields()
                if (this.allCheckDigitsAreCorrect())
                    this.checkDigit = calcCheckDigit(this.joinFields())
            }
        }
        this.timing = Date.now() - timestamp
    }

    *calcRowValues() {
        const from = Math.round(this.width * .47)
        const to = Math.round(this.width * .53)
        for (const y of range(this.height))
            yield Math.round(average(map(x => this.getValue(x, y),
                                         range(from, to))))
    }

    getValue(x, y) {
        return this.rgba[
            1 + 4 * (x + this.xOffset + this.fullWidth * (y + this.yOffset))
        ] >> 2
    }

    findRows() {
        return ifLet(
            findBest(sumSizes, concatMap(
                threshold => this.getRowCandidates(
                    calcCollisions(this.rowValues, threshold)),
                new Set(this.rowValues))),
            adjustSegments)
    }

    *getRowCandidates(segments) {
        for (const [before, [row1, row2, row3], rowSizes, sepSizes, after]
             of getSegmentGroups(3, segments)) {
            if (rowSizes.some(x => x < this.minRowSize))
                continue

            const maxRowSize = Math.max(...rowSizes)
            const middle = (row3.position + row3.size - row1.position) / 2
                  + row1.position

            if (row2.position < middle
                && row2.position + row2.size > middle
                && maxRowSize > Math.max(...sepSizes)
                && maxRowSize < before.size
                && maxRowSize < after.size
                && rowsHaveSimilarSize(rowSizes)
                && rowsHaveSimilarSize(sepSizes))
                yield [row1, row2]
        }
    }

    findColumns() {
        return ifLet(
            findBest(
                columns => [-columns[0].position, sumSizes(columns)],
                concatMap(
                    threshold => this.getColumnCandidates(
                        calcCollisions(this.columnValues, threshold)),
                    new Set(this.columnValues))),
            adjustSegments)
    }

    *calcColumnValues () {
        for (const x of range(this.width)) {
            const values = Array.from(
                map(y => this.getValue(x, y),
                    range(this.rows[0].position,
                          this.rows[1].position + this.rows[1].size)))
            yield Math.min(...values) - Math.max(...values)
        }
    }

    *getColumnCandidates(segments) {
        for (const [before, cols, colSizes, sepSizes, after]
             of getSegmentGroups(NUMBER_OF_COLUMNS, segments))
            if (before.size > Math.max(...sepSizes)
                && colSizes.every(x => x >= this.minColumnSize)
                && columnsHaveSimilarSize(colSizes)
                && columnsHaveSimilarSize(sepSizes))
                yield cols
    }

    getFormat() {
        return this.getChar(
            DNIE_LAST_DIGIT_COLUMN, NUMBERS_ROW, DIGITS_OR_NULL
        ) == '<' ? DNI_FORMAT : DNIE_FORMAT
    }

    getFields() {
        return this.format.FIELD_GETTERS.concat(
            ['getBirthdate', 'getExpirationDate']
        ).map(getter => this[getter]())
    }

    joinFields() {
        return this.fields.map(
            field => field.value + field.checkDigit
        ).join('')
    }

    allCheckDigitsAreCorrect() {
        return this.fields.every(
            field => field.checkDigit == field.expectedCheckDigit)
    }

    getSerialNumber() {
        const value = this.getString(
            DNIE_SERIAL_NUMBER_COLUMN, NUMBERS_ROW, DNIE_SERIAL_NUMBER_LEN)
        return {value: value,
                checkDigit: this.getChar(
                    DNIE_SERIAL_NUMBER_CHECK_DIGIT_COLUMN, NUMBERS_ROW),
                expectedCheckDigit: calcCheckDigit(value)}
    }

    getOldNif() {
        const value = this.getString(
            DNI_NIF_COLUMN, NUMBERS_ROW, NIF_NUMBER_LEN)
        const letter = calcNifLetter(value)
        return {value: value,
                checkDigit: this.getString(
                    DNI_NIF_LETTER_COLUMN, NUMBERS_ROW, 2),
                expectedCheckDigit: letter + calcCheckDigit(value + letter)}
    }

    getNif() {
        const value = this.getString(
            DNIE_NIF_COLUMN, NUMBERS_ROW, NIF_NUMBER_LEN)
        return {value: value,
                checkDigit: this.getChar(DNIE_NIF_LETTER_COLUMN, NUMBERS_ROW),
                expectedCheckDigit: calcNifLetter(value)}
    }

    getBirthdate() {
        const value = this.getString(BIRTHDATE_COLUMN, DATES_ROW, DATE_LEN)
        return {value: value,
                checkDigit: this.getChar(
                    BIRTHDATE_CHECK_DIGIT_COLUMN, DATES_ROW),
                expectedCheckDigit: calcCheckDigit(value)}
    }

    getExpirationDate() {
        const value = this.getString(
            EXPIRATION_DATE_COLUMN, DATES_ROW, DATE_LEN)
        return {value: value,
                checkDigit: this.getChar(
                    EXPIRATION_DATE_CHECK_DIGIT_COLUMN, DATES_ROW),
                expectedCheckDigit: calcCheckDigit(value)}
    }

    getString(column, row, size) {
        return join(map(i => this.getChar(i + column, row),
                        range(size)))
    }

    getCharData(column, row) {
        const threshold = this.calcOptimalCharThreshold(column, row)
        const xStart = this.getCharXStart(column, row, threshold)
        const yStart = this.getCharYStart(column, row, threshold)
        const width  = this.getCharXEnd(column, row, threshold) - xStart + 1
        const height = this.getCharYEnd(column, row, threshold) - yStart + 1
        return {threshold: threshold, xStart: xStart, yStart: yStart,
                width: width, height: height}
    }

    getChar(column, row, chars) {
        const charData = this.getCharData(column, row)
        return findBest(
            c => this.templateScore(column, row, charData, this.templates[c]),
            chars || this.getPossibleChars(column, row)
        ) || '?'
    }

    getPossibleChars(column, row) {
        if (row == NUMBERS_ROW)
            return this.format.getPossibleChars(column)
        return DIGITS
    }

    isSignificantChar(column, row) {
        return (row == NUMBERS_ROW && column >= NUMBERS_COLUMN
                && this.format.isSignificantChar(column))
            || (row == DATES_ROW
                && ((column >= BIRTHDATE_COLUMN
                     && column <= BIRTHDATE_CHECK_DIGIT_COLUMN)
                    || (column >= EXPIRATION_DATE_COLUMN
                        && column <= EXPIRATION_DATE_CHECK_DIGIT_COLUMN)))
    }

    calcOptimalCharThreshold(column, row) {
        return .7 * average(concatMap(
            y => map(x => this.getCharValue(column, row, x, y),
                     range(this.columns[column].size)),
            range(this.rows[row].size)))
    }

    getCharXStart(column, row, threshold) {
        for (const x of range(this.columns[column].size))
            for (const y of range(this.rows[row].size))
                if (this.getCharBit(column, row, x, y, threshold))
                    return x
        return 0
    }

    getCharXEnd(column, row, threshold) {
        for (const x of reverseRange(this.columns[column].size))
            for (const y of range(this.rows[row].size))
                if (this.getCharBit(column, row, x, y, threshold))
                    return x
        return this.columns[column].size - 1
    }

    getCharYStart(column, row, threshold) {
        for (const y of range(this.rows[row].size))
            for (const x of range(this.columns[column].size))
                if (this.getCharBit(column, row, x, y, threshold))
                    return y
        return 0
    }

    getCharYEnd(column, row, threshold) {
        for (const y of reverseRange(this.rows[row].size))
            for (const x of range(this.columns[column].size))
                if (this.getCharBit(column, row, x, y, threshold))
                    return y
        return this.rows[row].size - 1
    }

    templateScore(column, row, charData, template) {
        return sum(concatMap(
            x => map(
                y => this.charBitMatchesTemplate(
                    column, row, x, y, charData, template
                ) ? 1 : 0,
                range(charData.height)),
            range(charData.width)))
    }

    charBitMatchesTemplate(column, row, x, y, charData, template) {
        return getTemplateBit(template, x, y, charData)
            == this.getRelativeCharBit(column, row, x, y, charData)
    }

    getRelativeCharBit(column, row, x, y, charData) {
        return this.getCharBit(column, row, x + charData.xStart,
                               y + charData.yStart, charData.threshold)
    }

    getCharBit(column, row, x, y, threshold) {
        return this.getCharValue(column, row, x, y) < threshold
    }

    getCharValue(column, row, x, y) {
        return this.getValue(x + this.columns[column].position,
                             y + this.rows[row].position)
    }
}

function rowsHaveSimilarSize(segments) {
    return coefficientOfVariation(segments) < .12
}

function columnsHaveSimilarSize(segments) {
    return coefficientOfVariation(segments) < .2
}

function calcNifLetter(cs) {
    return NIF_LETTERS[reduce((n, c) => n * 10 + c.charCodeAt() - CHAR_CODE_0,
                              0, cs)
                       % NIF_LETTERS.length]
}

function calcCheckDigit(cs) {
    return String.fromCharCode(
        CHAR_CODE_0 + sum(
            map(([c, w]) => w * (c.charCodeAt()
                                 - (c >= '0' && c <= '9'
                                    ? CHAR_CODE_0 : CHAR_CODE_A)),
                zip(cs, cycle([7, 3, 1])))
        ) % 10)
}

function getVideoImageData(video) {
    if (innerWidth > innerHeight && video.videoWidth && video.videoHeight)
        return getImageData(video, video.videoWidth, video.videoHeight)
}

async function loadImageData(url) {
    const img = await loadImage(url)
    return getImageData(img, img.width, img.height)
}

function loadImage(url) {
    const img = new Image
    img.crossOrigin = 'anonymous'
    img.src = url
    return new Promise((resolve, reject) => {
        img.addEventListener('load',  () => resolve(img))
        img.addEventListener('error', () => reject())
    })
}

function getImageData(element, width, height) {
    const canvas = document.createElement('canvas')
    canvas.width = width
    canvas.height = height
    const ctx = canvas.getContext('2d')
    ctx.drawImage(element, 0, 0, width, height)
    return ctx.getImageData(0, 0, width, height)
}

function sumSizes(xs) {
    return sum(getSizes(xs))
}

function findBest(getScore, items) {
    var bestItem, bestScore
    for (const item of items) {
        var itemScore = getScore(item)
        if (!Array.isArray(itemScore))
            itemScore = [itemScore]

        if (bestItem === undefined || isBetter(itemScore, bestScore)) {
            bestItem = item
            bestScore = itemScore
        }
    }
    return bestItem
}

function isBetter(xs, ys) {
    for (const i in xs) {
        if (xs[i] > ys[i])
            return true
        if (xs[i] < ys[i])
            return false
    }
    return false
}

function* calcCollisions(values, threshold) {
    var segment
    for (const i in values) {
        const isCollision = values[i] >= threshold
        if (!segment || segment.isCollision != isCollision) {
            if (segment && (segment.position > 0 || segment.isCollision))
                yield segment
            segment = {isCollision: isCollision,
                       position: parseInt(i),
                       size: 1}
        } else
            segment.size++
    }
    yield segment
}

function* getSegmentGroups(n, segments) {
    for (const nSegments of nWise(n * 2 + 1, segments, 2)) {
        const get = offset => map(i => nSegments[1 + offset + i * 2],
                                  range(n - offset))

        const items = Array.from(get(0))
        yield [nSegments[0], items, getSizesArray(items),
               getSizesArray(get(1)), nSegments[nSegments.length - 1]]
    }
}

function getSizes(xs) {
    return map(x => x.size, xs)
}

var getSizesArray = returnsArray(getSizes)

var adjustSegments = returnsArray(function* (segments) {
    const f = (a, b, c, d) => a
          ? (b.position - (c.position + c.size)) >> 1 : d
    for (const [prev, curr, next] of nWise(3, concat([0], segments, [0]))) {
        const n = f(next, next, curr, 0)
        const m = f(prev, curr, prev, n)
        yield {position: curr.position - m,
               size: curr.size + m + (n || m)}
    }
})

function getTemplateBit(template, x, y, charData) {
    return 127 > template.data[
        4 * (Math.round(x * template.width / charData.width)
             + template.width * Math.round(y * template.height
                                           / charData.height))]
}

function displayMessage(message) {
    document.getElementById('message').innerHTML = message
}

function clear() {
    document.getElementById('clearable').innerHTML = ''
}

function displayRows(ocrData) {
    for (const row of ocrData.rows)
        displayRow(ocrData, row)
}

function displayRow(ocrData, row) {
    const elt = createDivWithClass('row')
    setRow(ocrData, row, elt)
    document.getElementById('clearable').appendChild(elt)
}

function displaySignificantChars(ocrData) {
    ocrData.rows.forEach((row, r) =>
        ocrData.columns.forEach((column, c) => {
            if (ocrData.isSignificantChar(c, r))
                displaySignificantChar(ocrData, row, column)
        }))
}

function displaySignificantChar(ocrData, row, column) {
    const elt = createDivWithClass('char')
    setRow(ocrData, row, elt)
    setColumn(ocrData, column, elt)
    document.getElementById('clearable').appendChild(elt)
}

function createDivWithClass(cls) {
    const div = document.createElement('div')
    div.classList.add(cls)
    return div
}

function setRow({yOffset, ratio}, {position, size}, elt) {
    elt.style.top = (yOffset + position) / ratio + 'px'
    elt.style.height = size / ratio + 'px'
}

function setColumn({xOffset, ratio}, {position, size}, elt) {
    elt.style.left = (xOffset + position) / ratio + 'px'
    elt.style.width = size / ratio + 'px'
}

function displayCheckDigit(checkDigit) {
    document.getElementById('checkDigit').textContent = checkDigit
}

function hideOverlay() {
    document.getElementById('overlay').style.marginTop = '100vh'
}

function showOverlay() {
    document.getElementById('overlay').style.marginTop = 0
}

function returnsArray(f) {
    return (...args) => Array.from(f(...args))
}

function* nWise(n, xs, step) {
    xs = toArray(xs)
    for (const i of range(0, xs.length - (n - 1), step))
        yield xs.slice(i, i + n)
}

function* reverseRange(n) {
    for (var i = n - 1; i >= 0; i--)
        yield i
}

function* range(a, z, step = 1) {
    if (z === undefined) {
        z = a
        a = 0
    }
    for (var i = a; i < z; i += step)
        yield i
}

function sum(xs) {
    var total = 0
    for (const x of xs)
        total += x
    return total
}

function* cycle(xs) {
    xs = toArray(xs)
    for (;;)
        for (const x of xs)
            yield x
}

function iter(xs) {
    return xs[Symbol.iterator]()
}

function reduce(f, initial, xs) {
    for (const x of xs)
        initial = f(initial, x)
    return initial
}

function* zip(...is) {
    if (is.length == 1)
        for (const x of is[0])
            yield [x]
    else if (is.length > 1) {
        const xs = iter(is.shift())
        const ys = zip(...is)
        for (;;) {
            const x = xs.next(); if (x.done) break
            const y = ys.next(); if (y.done) break
            yield [x.value, ...y.value]
        }
    }
}

function* map(f, xs) {
    for (const x of xs)
        yield f(x)
}

function* concat(...xss) {
    for (const xs of xss)
        for (const x of xs)
            yield x
}

function concatMap(f, xs) {
    return concat(...map(f, xs))
}

function coefficientOfVariation(xs) {
    const mean = average(xs)
    return Math.sqrt(average(xs.map(x => Math.pow(x - mean, 2)))) / mean
}

function average(xs) {
    xs = toArray(xs)
    return sum(xs) / xs.length
}

function join(xs, sep = '') {
    return toArray(xs).join(sep)
}

function toArray(xs) {
    return Array.isArray(xs) ? xs : Array.from(xs)
}

function ifLet(x, then) {
    if (x)
        return then(x)
}
